import math
from typing import Dict, Iterator, List

from eth_utils import to_int
from ethpm_types import BaseModel, HexBytes
from pydantic import Field, validator

from evm_trace.base import CallTreeNode
from evm_trace.enums import CALL_OPCODES, CallType


class TraceFrame(BaseModel):
    """
    A low-level data structure modeling a transaction trace frame
    from the Geth RPC ``debug_traceTransaction``.
    """

    pc: int
    """Program counter."""

    op: str
    """Opcode."""

    gas: int
    """Remaining gas."""

    gas_cost: int = Field(alias="gasCost")
    """The cost to execute this opcode."""

    depth: int
    """
    The number of external jumps away the initially called contract (starts at 0).
    """

    stack: List[HexBytes] = []
    """Execution stack."""

    memory: List[HexBytes] = []
    """Execution memory."""

    storage: Dict[HexBytes, HexBytes] = {}
    """Contract storage."""

    @validator("pc", "gas", "gas_cost", "depth", pre=True)
    def validate_ints(cls, value):
        return int(value, 16) if isinstance(value, str) else value


def get_calltree_from_geth_call_trace(data: Dict) -> CallTreeNode:
    """
    Creates a CallTreeNode from a given transaction call trace.

    Args:
        data (Dict): The response from ``debug_traceTransaction`` when using
          ``tracer=callTracer``.

    Returns:
        :class:`~evm_trace.base.CallTreeNode`: Call tree of transaction trace.
    """

    data = _validate_data_from_call_tracer(data)
    root = CallTreeNode(**data)

    def fix_depth(r: CallTreeNode):
        for c in r.calls:
            c.depth = r.depth + 1
            fix_depth(c)

    fix_depth(root)
    return root


def get_calltree_from_geth_trace(
    trace: Iterator[TraceFrame], show_internal: bool = False, **root_node_kwargs
) -> CallTreeNode:
    """
    Creates a CallTreeNode from a given transaction trace.

    Args:
        trace (Iterator[TraceFrame]): Iterator of transaction trace frames.
        show_internal (bool): Boolean whether to display internal calls.
          Defaults to ``False``.
        root_node_kwargs (dict): Keyword arguments passed to the root ``CallTreeNode``.

    Returns:
        :class:`~evm_trace.base.CallTreeNode`: Call tree of transaction trace.
    """

    return _create_node_from_call(
        trace=trace,
        show_internal=show_internal,
        **root_node_kwargs,
    )


def _extract_memory(offset: HexBytes, size: HexBytes, memory: List[HexBytes]) -> HexBytes:
    """
    Extracts memory from the EVM stack.

    Args:
        offset (HexBytes): Offset byte location in memory.
        size (HexBytes): Number of bytes to return.
        memory (List[HexBytes]): Memory stack.

    Returns:
        HexBytes: Byte value from memory stack.
    """

    size_int = to_int(size)

    if size_int == 0:
        return HexBytes("")

    offset_int = to_int(offset)

    # Compute the word that contains the first byte
    start_word = math.floor(offset_int / 32)
    # Compute the word that contains the last byte
    stop_word = math.ceil((offset_int + size_int) / 32)

    end_index = stop_word + 1
    byte_slice = b"".join(memory[start_word:end_index])
    offset_index = offset_int % 32

    # NOTE: Add 4 for the selector.

    end_bytes_index = offset_index + size_int
    return_bytes = byte_slice[offset_index:end_bytes_index]
    return HexBytes(return_bytes)


def _create_node_from_call(
    trace: Iterator[TraceFrame], show_internal: bool = False, **node_kwargs
) -> CallTreeNode:
    """
    Use specified opcodes to create a branching callnode
    https://www.evm.codes/
    """

    if show_internal:
        raise NotImplementedError()

    node = CallTreeNode(**node_kwargs)
    for frame in trace:
        if frame.op in [x.value for x in CALL_OPCODES]:
            # NOTE: Because of the different meanings in structLog style gas values,
            # gas is not set for nodes created this way.
            child_node_kwargs = {"address": frame.stack[-2][-20:], "depth": frame.depth}

            if frame.op == CallType.CALL.value:
                child_node_kwargs["call_type"] = CallType.CALL
                child_node_kwargs["value"] = int(frame.stack[-3].hex(), 16)
                child_node_kwargs["calldata"] = _extract_memory(
                    offset=frame.stack[-4], size=frame.stack[-5], memory=frame.memory
                )
            elif frame.op == CallType.DELEGATECALL.value:
                child_node_kwargs["call_type"] = CallType.DELEGATECALL
                child_node_kwargs["calldata"] = _extract_memory(
                    offset=frame.stack[-3], size=frame.stack[-4], memory=frame.memory
                )
            else:
                child_node_kwargs["call_type"] = CallType.STATICCALL
                child_node_kwargs["calldata"] = _extract_memory(
                    offset=frame.stack[-3], size=frame.stack[-4], memory=frame.memory
                )

            child_node = _create_node_from_call(
                trace=trace, show_internal=show_internal, **child_node_kwargs
            )
            node.calls.append(child_node)

        # TODO: Handle internal nodes using JUMP and JUMPI

        elif frame.op == CallType.SELFDESTRUCT.value:
            # TODO: Handle the internal value transfer
            node.selfdestruct = True
            break

        elif frame.op == "STOP":
            # TODO: Handle "execution halted" vs. gas limit reached
            break

        elif frame.op in ("RETURN", "REVERT") and not node.returndata:
            node.returndata = _extract_memory(
                offset=frame.stack[-1], size=frame.stack[-2], memory=frame.memory
            )
            # TODO: Handle "execution halted" vs. gas limit reached
            node.failed = frame.op == "REVERT"
            break

        # TODO: Handle invalid opcodes (`node.failed = True`)
        # NOTE: ignore other opcodes

    # TODO: Handle "execution halted" vs. gas limit reached

    return node


def _validate_data_from_call_tracer(data: Dict) -> Dict:
    # Handle renames
    if "receiver" in data:
        data["address"] = data.pop("receiver")
    elif "to" in data:
        data["address"] = data.pop("to")
    if "input" in data:
        data["calldata"] = data.pop("input")
    if "output" in data:
        data["returndata"] = data.pop("output")
    if "gasUsed" in data:
        data["gas_cost"] = data.pop("gasUsed")
    if "gas" in data:
        data["gas_limit"] = data.pop("gas")
    if "type" in data:
        data["call_type"] = data.pop("type")

    # Remove unneeded keys
    unneeded_keys = ("sender", "from")
    for key in unneeded_keys:
        if key in data:
            del data[key]

    # Handle sub calls
    def fix_call_calls(r):
        r["calls"] = [
            _validate_data_from_call_tracer(x) for x in r.get("calls", []) if isinstance(x, dict)
        ]

    fix_call_calls(data)
    return data



from typing import List, Optional

from ethpm_types import BaseModel, HexBytes
from pydantic import validator

from evm_trace.display import get_tree_display
from evm_trace.enums import CallType


class CallTreeNode(BaseModel):
    """
    A higher-level object modeling a node in an execution call tree.
    Used by both Geth-style and Parity-style low-level data structures.
    """

    call_type: CallType
    """The type of call."""

    address: HexBytes = HexBytes("")
    """The contract address of the call."""

    value: int = 0
    """The amount of value sent on the call."""

    depth: int = 0
    """
    The number of external jumps away the initially called contract (starts at 0).
    """

    gas_limit: Optional[int] = None
    """
    The total amount of gas available.
    """

    gas_cost: Optional[int] = None  # calculated from call starting and return
    """The cost to execute this opcode."""

    calldata: HexBytes = HexBytes("")
    """Transaction calldata (inputs)."""

    returndata: HexBytes = HexBytes("")
    """Transaction returndata (outputs)."""

    calls: List["CallTreeNode"] = []
    """The list of external sub-calls this call makes."""

    selfdestruct: bool = False
    """Whether this is a SELFDESTRUCT opcode or not."""

    failed: bool = False
    """Whether the call failed or not."""

    def __str__(self) -> str:
        try:
            return get_tree_display(self)
        except Exception as err:
            return f"CallTreeNode (display_err={err})"

    def __repr__(self) -> str:
        return str(self)

    def __getitem__(self, index: int) -> "CallTreeNode":
        return self.calls[index]

    @validator("calldata", "returndata", "address", pre=True)
    def validate_bytes(cls, value):
        return HexBytes(value) if isinstance(value, str) else value

    @validator("value", "depth", pre=True)
    def validate_ints(cls, value):
        if not value:
            return 0

        return int(value, 16) if isinstance(value, str) else value

    @validator("gas_limit", "gas_cost", pre=True)
    def validate_optional_ints(cls, value):
        return int(value, 16) if isinstance(value, str) else value



import copy
from typing import Dict, List, TypeVar

from evm_trace.base import CallTreeNode

ContractID = TypeVar("ContractID")
MethodID = TypeVar("MethodID")
GasReport = Dict[ContractID, Dict[MethodID, List[int]]]


def get_gas_report(calltree: CallTreeNode) -> GasReport:
    """
    Extracts a gas report object from a :class:`~evm_trace.base.CallTreeNode`.

    Args:
        calltree (:class:`~evm_trace.base.CallTreeNode`): call tree used for gas report.

    Returns:
        :class:`~evm_trace.gas.GasReport`: Gas report structure from a call tree.
    """
    report = {
        calltree.address: {calltree.calldata[:4]: [calltree.gas_cost] if calltree.gas_cost else []}
    }
    return merge_reports(report, *map(get_gas_report, calltree.calls))


def merge_reports(*reports: GasReport) -> GasReport:
    """
    Merge method for merging a list of gas reports and combining a list of gas costs.
    If given a single report, it only returns it.
    """
    reports_ls = list(reports)
    num_reports = len(reports_ls)
    if num_reports == 0:
        return {}

    elif num_reports == 1:
        return reports_ls[0]

    merged_report: GasReport = copy.deepcopy(reports_ls.pop(0))
    for report in reports_ls:
        for outer_key, inner_dict in report.items():
            if outer_key not in merged_report:
                merged_report[outer_key] = inner_dict
                continue

            for inner_key, inner_list in report[outer_key].items():
                if inner_key in merged_report[outer_key]:
                    merged_report[outer_key][inner_key].extend(inner_list)
                else:
                    merged_report[outer_key][inner_key] = inner_list

    return merged_report



from typing import Any, Dict, List, Optional, Union, cast

from pydantic import BaseModel, Field, validator

from evm_trace.base import CallTreeNode
from evm_trace.enums import CallType


class CallAction(BaseModel):
    gas: int
    """
    The amount of gas available for the action.
    """

    input: Optional[str] = None
    receiver: Optional[str] = Field(alias="to", default=None)
    sender: str = Field(alias="from")
    value: int
    # only used to recover the specific call type
    call_type: str = Field(alias="callType", repr=False)

    @validator("value", "gas", pre=True)
    def convert_integer(cls, v):
        return int(v, 16)


class CreateAction(BaseModel):
    gas: int
    """
    The amount of gas available for the action.
    """

    init: str
    value: int

    @validator("value", "gas", pre=True)
    def convert_integer(cls, v):
        return int(v, 16)


class SelfDestructAction(BaseModel):
    address: str
    balance: int

    @validator("balance", pre=True)
    def convert_integer(cls, v):
        return int(v, 16) if isinstance(v, str) else int(v)


class ActionResult(BaseModel):
    """
    A base class for various OP-code-specified actions
    in Parity ``trace_transaction`` output.
    """

    gas_used: int = Field(alias="gasUsed")
    """
    The amount of gas utilized by the action. It does *not*
    include the ``21,000`` base fee or the data costs of ``4``
    for gas per zero byte and ``16`` gas per non-zero byte.
    """

    @validator("gas_used", pre=True)
    def convert_integer(cls, v):
        return int(v, 16) if isinstance(v, str) else int(v)


class CallResult(ActionResult):
    """
    The result of CALL.
    """

    output: str


class CreateResult(ActionResult):
    """
    The result of CREATE.
    """

    address: str
    code: str


ParityTraceAction = Union[CreateAction, CallAction, SelfDestructAction]
ParityTraceResult = Union[CallResult, CreateResult]


class ParityTrace(BaseModel):
    error: Optional[str] = None
    action: ParityTraceAction
    block_hash: str = Field(alias="blockHash")
    call_type: CallType = Field(alias="type")
    result: Optional[ParityTraceResult] = None
    subtraces: int
    trace_address: List[int] = Field(alias="traceAddress")
    transaction_hash: str = Field(alias="transactionHash")

    @validator("call_type", pre=True)
    def convert_call_type(cls, v, values) -> CallType:
        if isinstance(values["action"], CallAction):
            v = values["action"].call_type
        value = v.upper()
        if value == "SUICIDE":
            value = "SELFDESTRUCT"

        return CallType(value)


class ParityTraceList(BaseModel):
    __root__: List[ParityTrace]

    # pydantic models with custom root don't have this by default
    def __iter__(self):
        return iter(self.__root__)

    def __getitem__(self, item):
        return self.__root__[item]


def get_calltree_from_parity_trace(
    traces: ParityTraceList,
    root: Optional[ParityTrace] = None,
    **root_kwargs,
) -> CallTreeNode:
    """
    Create a :class:`~evm_trace.base.CallTreeNode` from output models using the Parity approach
    (e.g. from the ``trace_transaction`` RPC).

    Args:
        traces (:class:~evm_trace.parity.ParityTraceList): The list of parity trace nodes,
          likely loaded from the response data from the ``trace_transaction`` RPC response.
        root (:class:`~evm_trace.parity.ParityTrace`): The root parity trace node. Optional, uses
          the first item by default.
        **root_kwargs: Additional kwargs to append to the root node. Useful for adding gas for
          reverted calls.

    Returns:
        :class:`~evm_trace.base.CallTreeNode`
    """
    root = root or traces[0]
    failed = root.error is not None
    node_kwargs: Dict[Any, Any] = {
        "call_type": root.call_type,
        "failed": failed,
    }

    if root.call_type == CallType.CREATE:
        create_action: CreateAction = cast(CreateAction, root.action)
        create_result: Optional[CreateResult] = (
            cast(CreateResult, root.result) if root.result is not None else None
        )

        node_kwargs.update(
            value=create_action.value,
            gas_limit=create_action.gas,
        )
        if create_result:
            node_kwargs.update(gas_used=create_result.gas_used, address=create_result.address)

    elif root.call_type in (
        CallType.CALL,
        CallType.DELEGATECALL,
        CallType.STATICCALL,
        CallType.CALLCODE,
    ):
        call_action: CallAction = cast(CallAction, root.action)
        call_result: Optional[CallResult] = (
            cast(CallResult, root.result) if root.result is not None else None
        )

        node_kwargs.update(
            address=call_action.receiver,
            value=call_action.value,
            gas_limit=call_action.gas,
            calldata=call_action.input,
        )
        # no result if the call has an error
        if call_result:
            node_kwargs.update(
                gas_cost=call_result.gas_used,
                returndata=call_result.output,
            )

    elif root.call_type == CallType.SELFDESTRUCT:
        selfdestruct_action: SelfDestructAction = cast(SelfDestructAction, root.action)
        node_kwargs.update(
            address=selfdestruct_action.address,
        )

    trace_list: List[ParityTrace] = list(traces)
    subtraces: List[ParityTrace] = [
        sub
        for sub in trace_list
        if len(sub.trace_address) == len(root.trace_address) + 1
        and sub.trace_address[:-1] == root.trace_address
    ]
    node_kwargs["calls"] = [get_calltree_from_parity_trace(traces, root=sub) for sub in subtraces]
    node_kwargs = {**node_kwargs, **root_kwargs}
    node = CallTreeNode.parse_obj(node_kwargs)
    return node



from .base import CallTreeNode
from .enums import CallType
from .geth import TraceFrame, get_calltree_from_geth_call_trace, get_calltree_from_geth_trace
from .parity import ParityTrace, ParityTraceList, get_calltree_from_parity_trace

__all__ = [
    "CallTreeNode",
    "CallType",
    "get_calltree_from_geth_trace",
    "get_calltree_from_geth_call_trace",
    "get_calltree_from_parity_trace",
    "ParityTrace",
    "ParityTraceList",
    "TraceFrame",
]



import pytest

from evm_trace.base import CallTreeNode
from evm_trace.enums import CallType

from .expected_traces import (
    CALL_TRACE_EXPECTED_OUTPUT,
    DELEGATECALL_TRACE_EXPECTED_OUTPUT,
    STATIC_TRACE_EXPECTED_OUTPUT,
)

EXPECTED_OUTPUT_MAP = {
    CallType.CALL: CALL_TRACE_EXPECTED_OUTPUT,
    CallType.STATICCALL: STATIC_TRACE_EXPECTED_OUTPUT,
    CallType.DELEGATECALL: DELEGATECALL_TRACE_EXPECTED_OUTPUT,
}


@pytest.fixture(scope="session")
def call_tree(call_tree_data):
    return CallTreeNode(**call_tree_data)


class TestCallTreeNode:
    def test_call_tree_validation_passes(self, call_tree_data):
        tree = CallTreeNode(**call_tree_data)
        assert tree

    def test_call_tree_mutable_representation(self, call_tree):
        expected = EXPECTED_OUTPUT_MAP[call_tree.call_type].strip()
        assert repr(call_tree) == expected



from enum import Enum


class CallType(Enum):
    INTERNAL = "INTERNAL"  # Non-opcode internal call
    CREATE = "CREATE"
    CREATE2 = "CREATE2"
    CALL = "CALL"
    DELEGATECALL = "DELEGATECALL"
    STATICCALL = "STATICCALL"
    CALLCODE = "CALLCODE"
    SELFDESTRUCT = "SELFDESTRUCT"


CALL_OPCODES = (CallType.CALL, CallType.CALLCODE, CallType.DELEGATECALL, CallType.STATICCALL)



from typing import TYPE_CHECKING, Iterator, Optional, cast

from eth_typing import ChecksumAddress
from eth_utils import to_checksum_address

if TYPE_CHECKING:
    from evm_trace.base import CallTreeNode


def get_tree_display(call: "CallTreeNode") -> str:
    return "\n".join([str(t) for t in TreeRepresentation.make_tree(call)])


class TreeRepresentation:
    FILE_MIDDLE_PREFIX = "├──"
    FILE_LAST_PREFIX = "└──"
    PARENT_PREFIX_MIDDLE = "    "
    PARENT_PREFIX_LAST = "│   "

    def __init__(
        self,
        call: "CallTreeNode",
        parent: Optional["TreeRepresentation"] = None,
        is_last: bool = False,
    ):
        self.call = call
        self.parent = parent
        self.is_last = is_last

    @property
    def depth(self) -> int:
        return self.call.depth

    @property
    def title(self) -> str:
        call_type = self.call.call_type.value
        address_hex_str = self.call.address.hex() if self.call.address else None

        try:
            address = to_checksum_address(address_hex_str) if address_hex_str else None
        except (ImportError, ValueError):
            # Ignore checksumming if user does not have eth-hash backend installed.
            address = cast(ChecksumAddress, address_hex_str)

        cost = self.call.gas_cost
        call_path = str(address) if address else ""
        if self.call.calldata:
            call_path = f"{call_path}." if call_path else ""
            call_path = f"{call_path}<{self.call.calldata[:4].hex()}>"

        call_path = (
            f"[reverted] {call_path}" if self.call.failed and self.parent is None else call_path
        )
        call_path = call_path.strip()
        node_title = f"{call_type}: {call_path}" if call_path else call_type
        if cost is not None:
            node_title = f"{node_title} [{cost} gas]"

        return node_title

    @classmethod
    def make_tree(
        cls,
        root: "CallTreeNode",
        parent: Optional["TreeRepresentation"] = None,
        is_last: bool = False,
    ) -> Iterator["TreeRepresentation"]:
        displayable_root = cls(root, parent=parent, is_last=is_last)
        yield displayable_root

        count = 1
        for child_node in root.calls:
            is_last = count == len(root.calls)
            if child_node.calls:
                yield from cls.make_tree(child_node, parent=displayable_root, is_last=is_last)
            else:
                yield cls(child_node, parent=displayable_root, is_last=is_last)

            count += 1

    def __str__(self) -> str:
        if self.parent is None:
            return self.title

        filename_prefix = self.FILE_LAST_PREFIX if self.is_last else self.FILE_MIDDLE_PREFIX

        parts = [f"{filename_prefix} {self.title}"]
        parent = self.parent
        while parent and parent.parent is not None:
            parts.append(self.PARENT_PREFIX_MIDDLE if parent.is_last else self.PARENT_PREFIX_LAST)
            parent = parent.parent

        return "".join(reversed(parts))



from typing import List

from ethpm_types import HexBytes

from evm_trace import CallTreeNode
from evm_trace.gas import GasReport, get_gas_report, merge_reports

# Simplified version of gas reports only for testing purposes
CONTRACT_A = HexBytes("0x0000000000000000000000000000000000000001")
CONTRACT_B = HexBytes("0x0000000000000000000000000000000000000002")
CONTRACT_C = HexBytes("0x0000000000000000000000000000000000000003")
METHOD_A = HexBytes("0x181d60da")
METHOD_B = HexBytes("0xc7cee1b7")
METHOD_C = HexBytes("0xd76d0659")

reports: List[GasReport] = [
    {
        CONTRACT_A: {METHOD_A: [100, 101, 100, 102]},
        CONTRACT_B: {METHOD_B: [200, 202, 202, 200, 200]},
    },
    {
        CONTRACT_A: {METHOD_A: [105, 106]},
        CONTRACT_B: {METHOD_A: [200, 201]},
        CONTRACT_C: {METHOD_C: [300]},
    },
]


def test_get_gas_report(call_tree_data):
    tree = CallTreeNode(**call_tree_data)
    gas_report = get_gas_report(tree)

    def assert_all(t):
        assert t.address in gas_report
        for c in t.calls:
            assert_all(c)

    assert_all(tree)


def test_merge_reports():
    merged = merge_reports(*reports)
    assert merged == {
        CONTRACT_A: {METHOD_A: [100, 101, 100, 102, 105, 106]},
        CONTRACT_B: {METHOD_A: [200, 201], METHOD_B: [200, 202, 202, 200, 200]},
        CONTRACT_C: {METHOD_C: [300]},
    }


def test_merge_single_report():
    actual = merge_reports(reports[0])
    expected = reports[0]
    assert actual == expected


def test_merge_no_reports():
    assert merge_reports() == {}



from pathlib import Path

import pytest

from evm_trace.parity import ParityTraceList, get_calltree_from_parity_trace

from .expected_traces import (
    PARITY_CALL_TRACE_EXPECTED_OUTPUT,
    PARITY_CREATE_REVERT_TRACE_EXPECTED_OUTPUT,
    PARITY_CREATE_TRACE_EXPECTED_OUTPUT,
    PARITY_OUT_OF_GAS_TRACE_EXPECTED_OUTPUT,
    PARITY_REVERT_TRACE_EXPECTED_OUTPUT,
    PARITY_REVERT_TRACE_WITH_MESSAGE_EXPECTED_OUTPUT,
    PARITY_SELFDESTRUCT_TRACE_EXPECTED_OUTPUT,
)

DATA_PATH = Path(__file__).parent / "data" / "parity"
EXPECTED_OUTPUT_MAP = {
    "call": PARITY_CALL_TRACE_EXPECTED_OUTPUT,
    "create": PARITY_CREATE_TRACE_EXPECTED_OUTPUT,
    "selfdestruct": PARITY_SELFDESTRUCT_TRACE_EXPECTED_OUTPUT,
    "revert": PARITY_REVERT_TRACE_EXPECTED_OUTPUT,
    "error": PARITY_OUT_OF_GAS_TRACE_EXPECTED_OUTPUT,
    "create_revert": PARITY_CREATE_REVERT_TRACE_EXPECTED_OUTPUT,
    "revert_with_message": PARITY_REVERT_TRACE_WITH_MESSAGE_EXPECTED_OUTPUT,
}


@pytest.mark.parametrize(
    "name",
    ("call", "create", "selfdestruct", "revert", "error", "create_revert", "revert_with_message"),
)
def test_parity(name):
    assert name in EXPECTED_OUTPUT_MAP, f"Missing expected output set for '{name}'."
    path = DATA_PATH / f"{name}.json"
    traces = ParityTraceList.parse_file(path)
    actual = repr(get_calltree_from_parity_trace(traces))

    expected = EXPECTED_OUTPUT_MAP[name].strip()
    assert actual == expected



from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional, Type, Union

from eth.vm.memory import Memory
from eth.vm.stack import Stack
from eth_utils import to_checksum_address
from hexbytes import HexBytes
from msgspec import Struct
from msgspec.json import Decoder

# opcodes grouped by the number of items they pop from the stack
# fmt: off
POP_OPCODES = {
    1: ["EXTCODEHASH", "ISZERO", "NOT", "BALANCE", "CALLDATALOAD", "EXTCODESIZE", "BLOCKHASH", "POP", "MLOAD", "SLOAD", "JUMP", "SELFDESTRUCT"],  # noqa: E501
    2: ["SHL", "SHR", "SAR", "REVERT", "ADD", "MUL", "SUB", "DIV", "SDIV", "MOD", "SMOD", "EXP", "SIGNEXTEND", "LT", "GT", "SLT", "SGT", "EQ", "AND", "XOR", "OR", "BYTE", "SHA3", "MSTORE", "MSTORE8", "SSTORE", "JUMPI", "RETURN"],  # noqa: E501
    3: ["RETURNDATACOPY", "ADDMOD", "MULMOD", "CALLDATACOPY", "CODECOPY", "CREATE"],
    4: ["CREATE2", "EXTCODECOPY"],
    6: ["STATICCALL", "DELEGATECALL"],
    7: ["CALL", "CALLCODE"]
}
# fmt: on
POPCODES = {op: n for n, opcodes in POP_OPCODES.items() for op in opcodes}
POPCODES.update({f"LOG{n}": n + 2 for n in range(0, 5)})
POPCODES.update({f"SWAP{i}": i + 1 for i in range(1, 17)})
POPCODES.update({f"DUP{i}": i for i in range(1, 17)})


class uint256(int):
    pass


class VMTrace(Struct):
    code: HexBytes
    """The code to be executed."""
    ops: List[VMOperation]
    """The operations executed."""


class VMOperation(Struct):
    pc: int
    """The program counter."""
    cost: int
    """The gas cost for this instruction."""
    ex: Optional[VMExecutedOperation]
    """Information concerning the execution of the operation."""
    sub: Optional[VMTrace]
    """Subordinate trace of the CALL/CREATE if applicable."""
    op: str
    """Opcode that is being called."""
    idx: str
    """Index in the tree."""


class VMExecutedOperation(Struct):
    used: int
    """The amount of remaining gas."""
    push: List[HexBytes]
    """The stack item placed, if any."""
    mem: Optional[MemoryDiff]
    """If altered, the memory delta."""
    store: Optional[StorageDiff]
    """The altered storage value, if any."""


class MemoryDiff(Struct):
    off: int
    """Offset into memory the change begins."""
    data: HexBytes
    """The changed data."""


class StorageDiff(Struct):
    key: uint256
    """Which key in storage is changed."""
    val: uint256
    """What the value has been changed to."""


class VMTraceFrame(Struct):
    """
    A synthetic trace frame representing the state at a step of execution.
    """

    address: str
    pc: int
    op: str
    depth: int
    stack: List[int]
    memory: Union[bytes, memoryview]
    storage: Dict[int, int]


def to_address(value):
    # clear the padding and expand to 32 bytes
    return to_checksum_address(value[-20:].rjust(20, b"\x00"))


def to_trace_frames(
    trace: VMTrace,
    depth: int = 1,
    address: str = "",
    copy_memory: bool = True,
) -> Iterator[VMTraceFrame]:
    """
    Replays a VMTrace and yields trace frames at each step of the execution.
    Can be used as a much faster drop-in replacement for Geth-style traces.

    Args:
        trace (VMTrace): A decoded trace from a `trace_` rpc.
        depth (int): A depth of the call being processed. automatically populated.
        address (str): The address of the contract being executed. auto populated
            except the root call.
        copy_memory (bool): Whether to copy memory when returning trace frames.
            Disable for a speedup when dealing with traces using a large amount of memory.
            when disabled, `VMTraceFrame.memory` becomes `memoryview` instead of `bytes`, which
            works like a pointer at the memory `bytearray`. this means you must process the
            frames immediately, otherwise you risk memory value mutating further into execution.

    Returns:
        Iterator[VMTraceFrame]: An iterator of synthetic traces which can be used as a drop-in
        replacement for Geth-style traces. also contains the address of the current contract
        context.
    """
    memory = Memory()
    stack = Stack()
    storage: Dict[int, int] = {}
    call_address = ""
    read_memory = memory.read_bytes if copy_memory else memory.read

    for op in trace.ops:
        if op.ex and op.ex.mem:
            memory.extend(op.ex.mem.off, len(op.ex.mem.data))

        # geth convention is to return after memory expansion, but before the operation is applied
        yield VMTraceFrame(
            address=address,
            pc=op.pc,
            op=op.op,
            depth=depth,
            stack=[val for _, val in stack.values],
            memory=read_memory(0, len(memory)),
            storage=storage.copy(),
        )

        if op.op in ["CALL", "DELEGATECALL", "STATICCALL"]:
            call_address = to_address(stack.values[-2][1])

        if op.ex:
            if op.ex.mem:
                memory.write(op.ex.mem.off, len(op.ex.mem.data), op.ex.mem.data)

            num_pop = POPCODES.get(op.op)
            if num_pop:
                stack.pop_any(num_pop)

            for item in op.ex.push:
                stack.push_bytes(item)

            if op.ex.store:
                storage[op.ex.store.key] = op.ex.store.val

        if op.sub:
            yield from to_trace_frames(
                op.sub, depth=depth + 1, address=call_address, copy_memory=copy_memory
            )


class RPCResponse(Struct):
    result: Union[RPCTraceResult, List[RPCTraceResult]]


class RPCTraceResult(Struct):
    trace: Optional[List]
    vmTrace: VMTrace
    stateDiff: Optional[Dict]


def dec_hook(type: Type, obj: Any) -> Any:
    if type is uint256:
        return uint256(obj, 16)
    elif type is HexBytes:
        return HexBytes(obj)


def from_rpc_response(buffer: bytes) -> Union[VMTrace, List[VMTrace]]:
    """
    Decode structured data from a raw `trace_replayTransaction` or `trace_replayBlockTransactions`.
    """
    resp = Decoder(RPCResponse, dec_hook=dec_hook).decode(buffer)

    if isinstance(resp.result, list):
        return [i.vmTrace for i in resp.result]

    return resp.result.vmTrace



import pytest
from hexbytes import HexBytes
from pydantic import ValidationError

from evm_trace.enums import CallType
from evm_trace.geth import (
    TraceFrame,
    get_calltree_from_geth_call_trace,
    get_calltree_from_geth_trace,
)


class TestTraceFrame:
    def test_validation_passes(self, trace_frame_data):
        frame = TraceFrame(**trace_frame_data)
        assert frame

    def test_no_memory(self):
        pc = 0
        op = "REVERT"
        gas = 4732305
        gas_cost = 3
        depth = 1
        stack = []
        trace_frame = TraceFrame(pc=pc, op=op, gas=gas, gasCost=gas_cost, depth=depth, stack=stack)
        assert trace_frame.pc == pc
        assert trace_frame.op == op
        assert trace_frame.gas == gas
        assert trace_frame.gas_cost == gas_cost
        assert trace_frame.depth == depth
        assert trace_frame.stack == []

    @pytest.mark.parametrize(
        "test_data",
        (
            {"stack": ["potato"]},
            {"memory": ["potato"]},
            {"storage": {"piggy": "dippin"}},
        ),
    )
    def test_validation_fails(self, test_data, trace_frame_data):
        data = {**trace_frame_data, **test_data}
        with pytest.raises(ValidationError):
            TraceFrame(**data)


def test_get_calltree_from_geth_trace(trace_frame_data):
    trace_frame_data["op"] = "RETURN"
    returndata = HexBytes("0x0000000000000000000000004d4d2c55eae97a04acafb66011df29463b665732")
    root_node_kwargs = {
        "gas_cost": 123,
        "gas_limit": 1234,
        "address": "0x56764a0000000000000000000000000000000031",
        "calldata": HexBytes("0x21325"),
        "value": 34,
        "call_type": CallType.CALL,
        "failed": False,
        "returndata": returndata,
    }
    frames = (TraceFrame(**d) for d in (trace_frame_data,))
    actual = get_calltree_from_geth_trace(frames, **root_node_kwargs)

    # Tests against a bug where we could not set the return data.
    assert actual.returndata == returndata


def test_get_calltree_from_geth_call_trace(geth_call_trace_data):
    node = get_calltree_from_geth_call_trace(geth_call_trace_data)
    expected = """
    CALL: 0xF2Df0b975c0C9eFa2f8CA0491C2d1685104d2488.<0x372dca07> [280038 gas]
├── CALL: 0xBcF7FFFD8B256Ec51a36782a52D0c34f6474D951.<0x045856de> [34004 gas]
├── CALL: 0xBcF7FFFD8B256Ec51a36782a52D0c34f6474D951.<0xbeed0f85> [62443 gas]
│   ├── STATICCALL: 0x274b028b03A250cA03644E6c578D81f019eE1323.<0x7007cbe8> [1497 gas]
│   └── CALL: 0x274b028b03A250cA03644E6c578D81f019eE1323.<0x878fb701> [41458 gas]
├── CALL: 0xBcF7FFFD8B256Ec51a36782a52D0c34f6474D951.<0xb27b8804> [723 gas]
├── CALL: 0xBcF7FFFD8B256Ec51a36782a52D0c34f6474D951.<0xb9e5b20a> [91563 gas]
│   ├── STATICCALL: 0x274b028b03A250cA03644E6c578D81f019eE1323.<0xe5e1d93f> [2375 gas]
│   ├── CALL: 0x274b028b03A250cA03644E6c578D81f019eE1323.<0x878fb701> [33858 gas]
│   ├── CALL: 0x274b028b03A250cA03644E6c578D81f019eE1323.<0x90bb7141> [25171 gas]
│   └── CALL: 0x274b028b03A250cA03644E6c578D81f019eE1323.<0x90bb7141> [23171 gas]
├── STATICCALL: 0x274b028b03A250cA03644E6c578D81f019eE1323.<0xbff2e095> [2814 gas]
├── STATICCALL: 0xBcF7FFFD8B256Ec51a36782a52D0c34f6474D951.<0x9155fd57> [859 gas]
├── CALL: 0xBcF7FFFD8B256Ec51a36782a52D0c34f6474D951.<0xbeed0f85> [32643 gas]
│   ├── STATICCALL: 0x274b028b03A250cA03644E6c578D81f019eE1323.<0x7007cbe8> [1497 gas]
│   └── CALL: 0x274b028b03A250cA03644E6c578D81f019eE1323.<0x878fb701> [25758 gas]
└── CALL: 0xBcF7FFFD8B256Ec51a36782a52D0c34f6474D951.<0xbeed0f85> [32643 gas]
    ├── STATICCALL: 0x274b028b03A250cA03644E6c578D81f019eE1323.<0x7007cbe8> [1497 gas]
    └── CALL: 0x274b028b03A250cA03644E6c578D81f019eE1323.<0x878fb701> [25758 gas]
    """
    assert repr(node) == expected.strip()
